package com.sun.jmx.remote.internal;

import java.util.List;
import javax.management.NotificationBroadcaster;
import java.security.AccessControlContext;
import javax.management.MBeanPermission;
import javax.management.ObjectInstance;
import javax.security.auth.Subject;
import javax.management.Notification;
import java.util.Iterator;
import javax.management.MBeanServerNotification;
import java.util.ArrayList;
import javax.management.MBeanServerDelegate;
import javax.management.remote.TargetedNotification;
import javax.management.remote.NotificationResult;
import javax.management.ListenerNotFoundException;
import java.util.Collections;
import java.util.HashSet;
import javax.management.MalformedObjectNameException;
import java.io.IOException;
import java.security.PrivilegedActionException;
import java.security.AccessController;
import javax.management.InstanceNotFoundException;
import java.security.PrivilegedExceptionAction;
import javax.management.NotificationFilter;
import com.sun.jmx.remote.util.EnvHelp;
import java.util.HashMap;
import com.sun.jmx.remote.util.ClassLogger;
import com.sun.jmx.remote.security.NotificationAccessController;
import java.util.Set;
import javax.management.ObjectName;
import java.util.Map;
import javax.management.MBeanServer;

public class ServerNotifForwarder
{
    private final ServerNotifForwarder.NotifForwarderBufferFilter bufferFilter;
    private MBeanServer mbeanServer;
    private final String connectionId;
    private final long connectionTimeout;
    private static int listenerCounter;
    private static final int[] listenerCounterLock;
    private NotificationBuffer notifBuffer;
    private final Map<ObjectName, Set<ServerNotifForwarder.IdAndFilter>> listenerMap;
    private boolean terminated;
    private final int[] terminationLock;
    static final String broadcasterClass;
    private final boolean checkNotificationEmission;
    private final NotificationAccessController notificationAccessController;
    private static final ClassLogger logger;
    
    public ServerNotifForwarder(MBeanServer mbeanServer, Map<String, ?> map, NotificationBuffer notifBuffer, String connectionId) {
        this.bufferFilter = new ServerNotifForwarder.NotifForwarderBufferFilter();
        this.listenerMap = new HashMap();
        this.terminated = false;
        this.terminationLock = new int[0];
        this.mbeanServer = mbeanServer;
        this.notifBuffer = notifBuffer;
        this.connectionId = connectionId;
        this.connectionTimeout = EnvHelp.getServerConnectionTimeout(map);
        this.checkNotificationEmission = EnvHelp.computeBooleanFromString((String)map.get("jmx.remote.x.check.notification.emission"));
        this.notificationAccessController = EnvHelp.getNotificationAccessController(map);
    }
    
    public Integer addNotificationListener(final ObjectName instance, NotificationFilter notificationFilter) throws InstanceNotFoundException, IOException {
        if (ServerNotifForwarder.logger.traceOn()) {
            ServerNotifForwarder.logger.trace("addNotificationListener", "Add a listener at " + instance);
        }
        this.checkState();
        this.checkMBeanPermission(instance, "addNotificationListener");
        if (this.notificationAccessController != null) {
            this.notificationAccessController.addNotificationListener(this.connectionId, instance, this.getSubject());
        }
        try {
            if (!(boolean)AccessController.doPrivileged(new PrivilegedExceptionAction<Boolean>(this, instance) {
                
                @Override
                public Boolean run() throws InstanceNotFoundException {
                    return Boolean.valueOf(ServerNotifForwarder.this.mbeanServer.isInstanceOf(name, ServerNotifForwarder.broadcasterClass));
                }
            })) {
                throw new IllegalArgumentException("The specified MBean [" + instance + "] is not a NotificationBroadcaster object.");
            }
        }
        catch (PrivilegedActionException ex) {
            throw (InstanceNotFoundException)extractException(ex);
        }
        Integer listenerID = this.getListenerID();
        if (instance.getDomain() == null || instance.getDomain().equals("")) {
            try {
                instance = ObjectName.getInstance(this.mbeanServer.getDefaultDomain(), instance.getKeyPropertyList());
            }
            catch (MalformedObjectNameException cause) {
                IOException ex2 = new IOException(cause.getMessage());
                ex2.initCause(cause);
                throw ex2;
            }
        }
        synchronized (this.listenerMap) {
            ServerNotifForwarder.IdAndFilter o = new ServerNotifForwarder.IdAndFilter(listenerID, notificationFilter);
            Set singleton = (Set<ServerNotifForwarder.IdAndFilter>)this.listenerMap.get(instance);
            if (singleton == null) {
                singleton = Collections.singleton(o);
            }
            else {
                if (singleton.size() != 0) {
                    singleton = new HashSet(singleton);
                }
                singleton.add(o);
            }
            this.listenerMap.put(instance, singleton);
        }
        return listenerID;
    }
    
    public void removeNotificationListener(ObjectName obj, Integer[] array) throws Exception {
        if (ServerNotifForwarder.logger.traceOn()) {
            ServerNotifForwarder.logger.trace("removeNotificationListener", "Remove some listeners from " + obj);
        }
        this.checkState();
        this.checkMBeanPermission(obj, "removeNotificationListener");
        if (this.notificationAccessController != null) {
            this.notificationAccessController.removeNotificationListener(this.connectionId, obj, this.getSubject());
        }
        Exception ex = null;
        for (int i = 0; i < array.length; ++i) {
            try {
                this.removeNotificationListener(obj, array[i]);
            }
            catch (Exception ex2) {
                if (ex != null) {
                    ex = ex2;
                }
            }
        }
        if (ex != null) {
            throw ex;
        }
        return;
    }
    
    public void removeNotificationListener(ObjectName objectName, Integer obj) throws InstanceNotFoundException, ListenerNotFoundException, IOException {
        if (ServerNotifForwarder.logger.traceOn()) {
            ServerNotifForwarder.logger.trace("removeNotificationListener", "Remove the listener " + obj + " from " + objectName);
        }
        this.checkState();
        if (objectName != null && !objectName.isPattern() && !this.mbeanServer.isRegistered(objectName)) {
            throw new InstanceNotFoundException("The MBean " + objectName + " is not registered.");
        }
        synchronized (this.listenerMap) {
            Set<Object> set = (Set)this.listenerMap.get(objectName);
            ServerNotifForwarder.IdAndFilter idAndFilter = new ServerNotifForwarder.IdAndFilter(obj, null);
            if (set == null || !set.contains(idAndFilter)) {
                throw new ListenerNotFoundException("Listener not found");
            }
            if (set.size() == 1) {
                this.listenerMap.remove(objectName);
            }
            else {
                set.remove(idAndFilter);
            }
        }
        return;
    }
    
    public NotificationResult fetchNotifs(long lng, long n, int i) {
        if (ServerNotifForwarder.logger.traceOn()) {
            ServerNotifForwarder.logger.trace("fetchNotifs", "Fetching notifications, the startSequenceNumber is " + lng + ", the timeout is " + n + ", the maxNotifications is " + i);
        }
        long min = Math.min(this.connectionTimeout, n);
        NotificationResult fetchNotifications;
        try {
            fetchNotifications = this.notifBuffer.fetchNotifications(this.bufferFilter, lng, min, i);
            this.snoopOnUnregister(fetchNotifications);
        }
        catch (InterruptedException ex) {
            fetchNotifications = new NotificationResult(0L, 0L, new TargetedNotification[0]);
        }
        if (ServerNotifForwarder.logger.traceOn()) {
            ServerNotifForwarder.logger.trace("fetchNotifs", "Forwarding the notifs: " + fetchNotifications);
            return fetchNotifications;
        }
        return fetchNotifications;
    }
    
    private void snoopOnUnregister(NotificationResult notificationResult) {
        ArrayList list = null;
        synchronized (this.listenerMap) {
            Set<Object> c = (Set)this.listenerMap.get(MBeanServerDelegate.DELEGATE_NAME);
            if (c == null || c.isEmpty()) {
                return;
            }
            list = new ArrayList(c);
        }
        for (TargetedNotification targetedNotification : notificationResult.getTargetedNotifications()) {
            Integer listenerID = targetedNotification.getListenerID();
            Iterator<Object> iterator = list.iterator();
            while (iterator.hasNext()) {
                if (((ServerNotifForwarder.IdAndFilter)iterator.next()).id == listenerID) {
                    Notification notification = targetedNotification.getNotification();
                    if (notification instanceof MBeanServerNotification && notification.getType().equals("JMX.mbean.unregistered")) {
                        ObjectName mBeanName = ((MBeanServerNotification)notification).getMBeanName();
                        synchronized (this.listenerMap) {
                            this.listenerMap.remove(mBeanName);
                        }
                    }
                    else {
                        continue;
                    }
                }
                else {
                    continue;
                }
            }
        }
    }
    
    public void terminate() {
        if (ServerNotifForwarder.logger.traceOn()) {
            ServerNotifForwarder.logger.trace("terminate", "Be called.");
        }
        synchronized (this.terminationLock) {
            if (this.terminated) {
                return;
            }
            this.terminated = true;
            synchronized (this.listenerMap) {
                this.listenerMap.clear();
            }
        }
        if (ServerNotifForwarder.logger.traceOn()) {
            ServerNotifForwarder.logger.trace("terminate", "Terminated.");
        }
    }
    
    private Subject getSubject() {
        return Subject.getSubject(AccessController.getContext());
    }
    
    private void checkState() throws IOException {
        synchronized (this.terminationLock) {
            if (this.terminated) {
                throw new IOException("The connection has been terminated.");
            }
        }
    }
    
    private Integer getListenerID() {
        synchronized (ServerNotifForwarder.listenerCounterLock) {
            return Integer.valueOf(ServerNotifForwarder.listenerCounter++);
        }
    }
    
    public final void checkMBeanPermission(ObjectName objectName, String s) throws InstanceNotFoundException, SecurityException {
        checkMBeanPermission(this.mbeanServer, objectName, s);
    }
    
    static void checkMBeanPermission(final MBeanServer mBeanServer, final ObjectName objectName, String actions) throws InstanceNotFoundException, SecurityException {
        SecurityManager securityManager = System.getSecurityManager();
        if (securityManager != null) {
            AccessControlContext context = AccessController.getContext();
            ObjectInstance objectInstance;
            try {
                objectInstance = (ObjectInstance)AccessController.doPrivileged(new PrivilegedExceptionAction<ObjectInstance>() {
                    @Override
                    public ObjectInstance run() throws InstanceNotFoundException {
                        return mBeanServer.getObjectInstance(objectName);
                    }
                });
            }
            catch (PrivilegedActionException ex) {
                throw (InstanceNotFoundException)extractException(ex);
            }
            securityManager.checkPermission(new MBeanPermission(objectInstance.getClassName(), null, objectName, actions), context);
        }
    }
    
    private boolean allowNotificationEmission(ObjectName objectName, TargetedNotification targetedNotification) {
        try {
            if (this.checkNotificationEmission) {
                this.checkMBeanPermission(objectName, "addNotificationListener");
            }
            if (this.notificationAccessController != null) {
                this.notificationAccessController.fetchNotification(this.connectionId, objectName, targetedNotification.getNotification(), this.getSubject());
            }
            return true;
        }
        catch (SecurityException ex) {
            if (ServerNotifForwarder.logger.debugOn()) {
                ServerNotifForwarder.logger.debug("fetchNotifs", "Notification " + targetedNotification.getNotification() + " not forwarded: the caller didn't have the required access rights");
            }
            return false;
        }
        catch (Exception obj) {
            if (ServerNotifForwarder.logger.debugOn()) {
                ServerNotifForwarder.logger.debug("fetchNotifs", "Notification " + targetedNotification.getNotification() + " not forwarded: got an unexpected exception: " + obj);
            }
            return false;
        }
    }
    
    private static Exception extractException(Exception exception) {
        while (exception instanceof PrivilegedActionException) {
            exception = ((PrivilegedActionException)exception).getException();
        }
        return exception;
    }
    
    static boolean access$300(ServerNotifForwarder serverNotifForwarder, ObjectName objectName, TargetedNotification targetedNotification) {
        return serverNotifForwarder.allowNotificationEmission(objectName, targetedNotification);
    }
    
    static {
        ServerNotifForwarder.listenerCounter = 0;
        listenerCounterLock = new int[0];
        broadcasterClass = NotificationBroadcaster.class.getName();
        logger = new ClassLogger("javax.management.remote.misc", "ServerNotifForwarder");
    }
    
    final class NotifForwarderBufferFilter implements NotificationBufferFilter
    {
        @Override
        public void apply(List<TargetedNotification> list, ObjectName objectName, Notification notification) {
            ServerNotifForwarder.IdAndFilter[] array;
            synchronized (ServerNotifForwarder.this.listenerMap) {
                Set<Object> set = (Set)ServerNotifForwarder.this.listenerMap.get(objectName);
                if (set == null) {
                    ServerNotifForwarder.logger.debug("bufferFilter", "no listeners for this name");
                    return;
                }
                array = new ServerNotifForwarder.IdAndFilter[set.size()];
                set.toArray(array);
            }
            for (ServerNotifForwarder.IdAndFilter idAndFilter : array) {
                NotificationFilter filter = idAndFilter.getFilter();
                if (filter == null || filter.isNotificationEnabled(notification)) {
                    ServerNotifForwarder.logger.debug("bufferFilter", "filter matches");
                    TargetedNotification o = new TargetedNotification(notification, idAndFilter.getId());
                    if (ServerNotifForwarder.access$300(ServerNotifForwarder.this, objectName, o)) {
                        list.add(o);
                    }
                }
            }
        }
    }
    
    private static class IdAndFilter
    {
        private Integer id;
        private NotificationFilter filter;
        
        IdAndFilter(Integer id, NotificationFilter filter) {
            this.id = id;
            this.filter = filter;
        }
        
        Integer getId() {
            return this.id;
        }
        
        NotificationFilter getFilter() {
            return this.filter;
        }
        
        @Override
        public int hashCode() {
            return this.id.hashCode();
        }
        
        @Override
        public boolean equals(Object o) {
            return o instanceof ServerNotifForwarder.IdAndFilter && ((ServerNotifForwarder.IdAndFilter)o).getId().equals(this.getId());
        }
    }
}
